from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from kele.syntax.base_classes import Assertion

if TYPE_CHECKING:
    from collections.abc import Sequence

    from kele.config import KBConfig
    from kele.control.grounding_selector import GroundingFlatTermWithWildCardSelector
    from kele.equality import Equivalence
    from kele.syntax.base_classes import FACT_TYPE, Question
    from kele.syntax.external import SankuManagementSystem


# FIXME: 存入事实库和规则库的信息应当满足safe的条件，我们将在 #132 中对CompoundTerm加入generic且引入safe的约束，以完成更精细的类型标注
class FactBase:
    """Store and manage facts for the inference engine.

    :param initial_facts_or_dir_or_path: Initial facts or path to load them from.
    :param equivalence_handler: Equivalence manager for equality updates.
    :param term_selector: Selector used for grounding term candidates.
    :param sk_system_handler: External knowledge system handler.
    :param args: Knowledge-base configuration.
    """

    def __init__(self,
                 initial_facts_or_dir_or_path: Sequence[FACT_TYPE] | str,
                 equivalence_handler: Equivalence,
                 term_selector: GroundingFlatTermWithWildCardSelector,
                 sk_system_handler: SankuManagementSystem,
                 args: KBConfig,
                 ):
        if isinstance(initial_facts_or_dir_or_path, str):
            self.facts: set[FACT_TYPE] = self._read_facts(initial_facts_or_dir_or_path)
        else:
            self.facts = set(initial_facts_or_dir_or_path)

        for fact in self.facts:
            self._validate_fact_for_storage(fact, check_free_variables=True)

        self.equivalence_handler = equivalence_handler
        self.term_selector = term_selector
        self.sk_system_handler = sk_system_handler  # HACK: 三库系统在解题前，会执行一次initial_by_question更新事实库。考虑到
        # 事实库的初始化是一个独立环节，此函数的调用被放置到FactBase内，而非最外层的InferenceEngine内。虽然从模块拆分的角度上FactBase和
        # 三库系统是两个系统，可能会建议在main函数内完成二者的交互。
        self._args = args
        self.max_facts = self._args.fact_cache_size

        self.initial_sign = False  # 引擎希望先挑选FactBase中的部分事实进行初始化。因此在__init__阶段我们认为初始化未完成，为False。
        # 当初始化完成后、当前值为True时才应当进行后续其他操作
        self.cur_facts: set[FACT_TYPE] = set()

    def _read_facts(self, path: str) -> set[FACT_TYPE]:
        """Load facts from a directory or file.

        :param path: Directory or file path.
        :returns: Set of facts.
        :raise NotImplementedError: will be completed before 0.0.1
        """
        raise NotImplementedError

    def _add_or_not(self, fact: FACT_TYPE) -> bool:
        """Decide whether a fact should be added to the fact base."""

        # 1. 判断事实是否已存在
        if fact in self.facts:
            return False  # 已存在的事实不再加入

        # 2. 判断是否超过事实库大小上限
        if self.max_facts != -1 and len(self.facts) >= self.max_facts:  # noqa: SIM103  # 函数还可以继续扩充，
            # 不应当直接简化为return (self.max_facts is not None...)，因此注释掉SIM103
            return False  # TODO: 如果事实库已满，最好是根据启发式决定是否丢弃现有事实或丢弃库内事实但保留当前事实

        # 3. 不合需求没必要保留（依赖启发式/NN判断）

        return True

    def add_facts(
        self,
        facts: Sequence[FACT_TYPE],
        *,
        force_add: bool = False,
        check_free_variables: bool = False,
    ) -> list[FACT_TYPE]:
        """
        Add facts and update equivalence classes.

        :param facts: Facts to add.
        :param force_add: Force insertion even if heuristics reject it.
        :param check_free_variables: Validate that facts have no free variables.
        :returns: Facts that were actually added.
        """
        added_facts = []
        for fact in facts:
            self._validate_fact_for_storage(fact, check_free_variables=check_free_variables)

        if force_add:  # 冗余一点避免多次判断force_add
            for fact in facts:
                self.facts.add(fact)
                added_facts.append(fact)
        else:
            for fact in facts:
                if self._add_or_not(fact):
                    self.facts.add(fact)
                    added_facts.append(fact)

        self.cur_facts |= set(added_facts)
        self.equivalence_handler.update_equiv_class(added_facts)
        if added_facts:
            self.term_selector.update_terms(facts=added_facts)

        return added_facts

    @staticmethod
    def _validate_fact_for_storage(fact: FACT_TYPE, *, check_free_variables: bool) -> None:
        if not isinstance(fact, Assertion):
            raise TypeError(f"Fact {fact} is not an Assertion, which is not allowed in the fact base.")
        if check_free_variables and fact.free_variables:
            raise ValueError(f"Fact {fact} contains free variables, which is not allowed.")

    def _select_initial_facts(self, question: Question, topn: int | None = None) -> list[FACT_TYPE]:
        """Select the most relevant facts for a question.

        :param question: Question being solved.
        :param topn: Limit on the number of facts to return; ``None`` disables the limit.
        :returns: Selected facts.
        """
        all_facts = list(self.facts)

        if topn is None or topn == -1 or topn >= len(all_facts):
            return all_facts

        return all_facts[:topn]  # TODO: 默认按插入顺序（或原始顺序）截取，是优化点

    def initial_fact_base(self, question: Question, topn: int | None = None) -> None:
        """Initialize the active fact set for a new question.

        :param question: Question being solved.
        :param topn: Optional cap on facts selected from the base.
        """
        selected_facts = self._select_initial_facts(question, topn)
        self.equivalence_handler.update_equiv_class(selected_facts)
        self.cur_facts |= set(selected_facts)

        sk_facts = self.sk_system_handler.initial_by_question(question, topn)
        self.add_facts(sk_facts)

        self.cur_facts |= set(sk_facts)
        self.initial_sign = True

    def reset_fact_base(self) -> None:
        """Reset the fact base to its pre-initialized state."""
        self.initial_sign = False
        self.cur_facts.clear()
        self.equivalence_handler.clear()  # 理论上如果这个函数仅有main调用，这一行是不应当有的。但我担心由于函数是外部的，如果被
        # 凑巧谁的代码调用过时，一步clear的冗余可以减少风险

    def get_facts(self) -> list[FACT_TYPE]:
        """Return facts used or derived up to the current round."""
        if self.initial_sign:
            return list(self.cur_facts)

        warnings.warn("Fact base has not been initialized yet.", stacklevel=2)
        return list(self.facts)

    def __str__(self) -> str:
        fact_count = len(self.facts)
        show_topn = 5

        fact_summary = ', '.join(str(fact) for fact in list(self.facts)[:show_topn])  # 只展示前五条事实

        # 如果事实数目大于5，提示用户后续还有更多
        if fact_count > show_topn:
            fact_summary += "..."

        return f"FactBase with {fact_count} facts. First 5 facts: [{fact_summary}]"



from collections.abc import Callable

from pydantic import BaseModel, ConfigDict

from .base_classes import FACT_TYPE, Assertion, Question


class SankuManagementSystem(BaseModel):
    """三库系统的抽象，主要用于类型检查和说明数据结构"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    knowledge: list[FACT_TYPE] = []
    initial_by_question: Callable[[Question, int | None], list[FACT_TYPE]] = lambda x, y: []
    # TODO(lbq): 需与泳融沟通，使用一个question和一个取出的上限值，由三库系统决定返回哪些有用的信息协助初始化。
    # 另外 HACK: 实则应该用memcached语言传入，而且这是一个http链接，不是函数
    query_assertion: Callable[[Assertion | list[Assertion]], list[Assertion]] = lambda x: []
    update_facts: Callable[[FACT_TYPE | list[FACT_TYPE]], None] = lambda x: None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from kele.syntax import CompoundTerm, Constant, Rule, Variable

if TYPE_CHECKING:
    from collections.abc import Generator, Mapping

    from kele.config import Config
    from kele.grounder import GroundedRule
    from kele.syntax import FACT_TYPE
    from kele.syntax.base_classes import _QuestionRule

    from ._tftable import TfTables

logger = logging.getLogger(__name__)


class _RuleNode:
    """进入一个RuleNode节点表明当前规则已被满足，需要进行新事实的生成"""

    def __init__(self, content: Rule, grounded_rule: GroundedRule, args: Config) -> None:
        self.content: Rule = content
        self.grounded_rule: GroundedRule = grounded_rule
        self._args = args

        self.left_table: TfTables  # RuleNode的父节点只会有一个，我们默认为left_table来统一名称

    def __str__(self) -> str:
        return str(self.content)

    def exec_check(self) -> list[FACT_TYPE]:
        """
        返回所有的新事实，用于判断终止条件和更新事实库
        :returns list[FACT_TYPE]: 新事实
        """
        return self._gen_new_facts()

    def _gen_new_facts(self) -> list[FACT_TYPE]:
        """FIXME: 这里要避免重复，可能是在grounding阶段就尽量移除掉重复的取值？不过此时的合并方式影响中途的去重"""
        new_facts: list[FACT_TYPE] = []
        if not self.left_table.true:
            return new_facts

        total_true_table = self.left_table.true

        for combination in total_true_table.iter_rows():
            new_fact = self.content.head.replace_variable(combination)
            new_facts.append(new_fact)

            if self._args.run.trace:
                self._trace(new_fact, combination)

        if self._args.executor.anti_join_used_facts:
            self.grounded_rule.receive_true_table(total_true_table)

        return new_facts

    @staticmethod
    def get_all_children() -> Generator[None]:
        """
        Yields:
            None: 因为RuleNode没有子节点，所以返回None
        """
        yield None

    def reset(self) -> None:
        if hasattr(self, 'left_table'):
            del self.left_table

    def _trace(self, new_fact: FACT_TYPE, combination: Mapping[Variable, Constant | CompoundTerm]) -> None:
        new_rule = Rule.from_parts(
            head=new_fact,
            body=self.content.body.replace_variable(combination),
            priority=self.content.priority,
        )

        for head in new_rule.head_units:
            self.grounded_rule.inference_path.add_infer_edge(consequent=head,
                                                             antecedents=new_rule.body_units,
                                                             grounded_rule=self.content.replace_variable(combination))  # FIXME: 这里
            # 存一整个grounded rule的instance，是不是有点太重了。而如果只是存一个实例化的Rule instance，那好像又不需要传


class _QuestionRuleNode(_RuleNode):
    """问题规则节点：额外负责收集查询解"""

    def __init__(self, content: _QuestionRule, grounded_rule: GroundedRule, args: Config) -> None:
        super().__init__(content, grounded_rule, args)
        self.question_rule = content
        self.solutions: list[Mapping[Variable, Constant | CompoundTerm]] = []

    def _gen_new_facts(self) -> list[FACT_TYPE]:  # TODO: 考虑与rule_node中的_gen_new_facts合并
        """生成新事实的同时收集查询解"""
        new_facts: list[FACT_TYPE] = []
        if not self.left_table.true:
            return new_facts

        total_true_table = self.left_table.true
        new_fact = self.content.head
        new_facts.append(new_fact)

        for combination in total_true_table.iter_rows():
            self.solutions.append(combination)

            if self._args.run.trace:
                self._trace(new_fact, combination)

        if self._args.executor.anti_join_used_facts:
            self.grounded_rule.receive_true_table(total_true_table)

        return new_facts

    def reset(self) -> None:
        """重写 reset：额外清空 solutions"""
        super().reset()
        self.solutions.clear()



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, cast, overload

from kele.knowledge_bases.builtin_base.builtin_facts import true_const
from kele.syntax import Constant, Variable
from kele.syntax.base_classes import Assertion, CompoundTerm, Formula, Question, Rule, _QuestionRule
from kele.syntax.dnf_converter import RuleSafetyProcesser

if TYPE_CHECKING:
    from collections.abc import Sequence

    from kele.config import KBConfig


class RuleBase:
    """Store and manage abstract rules for inference.

    :param rules_or_dir_or_path: Rules or path to load them from.
    :param args: Knowledge-base configuration.
    """

    def __init__(self, rules_or_dir_or_path: Sequence[Rule] | str, args: KBConfig):
        if isinstance(rules_or_dir_or_path, str):
            self.rules = list(dict.fromkeys(self._read_rules(rules_or_dir_or_path)))
            # 使用list代替set，以保持输入的规则顺序、因为set后的随机顺序会引入
            # 不必要的不确定性，随意的排序也对于推理无益
        else:
            self.rules = list(dict.fromkeys(rules_or_dir_or_path))  # 加入去重操作

        self.rule_suffix: int = 1

        self._var_counter = 0

        new_rules: list[Rule] = []
        for single_rule in self.rules:
            new_rules.extend(self._preprocess_rule(single_rule))
        self.rules = new_rules

        self.initial_sign: bool = False  # initial_rule_base后为True，和Fact base一样
        self.cur_rules: list[Rule] = []  # FactBase由于经常需要更新而用了set，而abstract rule不变，为了便捷用了list

        self.question_rule: list[_QuestionRule] = []  # 用于存储question转化成的rule，不干涉正常流程，不参与select等。

        self._args = args

    def _read_rules(self, path: str) -> list[Rule]:
        """Load rules from a directory or file.

        :param path: Directory or file path.
        :returns: Loaded rules.
        :raise NotImplementedError: will be completed before 0.0.1
        """
        raise NotImplementedError

    def add_rule(self, rule: Rule) -> None:
        """Add a rule at runtime (currently not supported)."""
        warnings.warn(
            "Not supported yet. When runtime rule additions are enabled, SelectionStrategy must add add_rule and "
            "adjust RuleCheckRule behavior as needed.",
            stacklevel=2,
        )
        if rule not in self.rules:
            self.rules.extend(self._preprocess_rule(rule))

    def get_rules(self) -> list[Rule]:
        """Return the active rules for the current run."""
        if self.initial_sign:
            return list(self.cur_rules)

        warnings.warn("Rule base has not been initialized yet.", stacklevel=2)
        return [r for r in self.rules if not isinstance(r, _QuestionRule)]

    def get_question_rules(self) -> list[_QuestionRule]:
        """Return rules derived from the current question."""
        return self.question_rule

    def _select_rules_from_current_question(self, question: Question, topn: int | None = None) -> list[Rule]:
        """
        Select rules based on the current question.

        :param question: Current question.
        :param topn: Optional limit on the number of rules to return.
        :returns: Selected rules.
        """
        # 由于选前k个而需要list或tuple一下  TODO: 另外：暂时将所有规则视为可能相关的规则，以后可以添加过滤逻辑或排序什么的。
        all_rules = list(self.rules)

        # 如果没有指定限制则返回所有规则。
        if topn is None or topn == -1 or topn >= len(all_rules):
            return all_rules

        return all_rules[:topn]

    def initial_rule_base(self, question: Question, topn: int | None = None) -> None:
        """Initialize the active rule set for a new question.

        :param question: Question being solved.
        :param topn: Optional cap on the number of selected rules.
        """
        if question.question:
            question_r = _QuestionRule.from_parts(head=Assertion.from_parts(_QuestionRule.QUESTION_SOLVED_FLAG, true_const),
                               body=question.question, name=_QuestionRule.QUESTIONRULE_NAME)
            self.question_rule.extend(self._preprocess_rule(question_r))

        # TODO: consider moving question rules handling to main.py to avoid mixing them into selector/base flow.
        selected_rules = self._select_rules_from_current_question(question, topn) + self.question_rule
        self.cur_rules = [rule for r in selected_rules for rule in self._preprocess_rule(r)]
        self.initial_sign = True

    def reset_rule_base(self) -> None:
        """
        Reset to the pre-initialized state.

        Clears active rules and marks the base as uninitialized.
        """
        self.cur_rules.clear()
        self.initial_sign = False

    def _preprocess_rule[T1: Rule | _QuestionRule](self, rule: T1) -> Sequence[T1]:
        return [self._resolve_variable_collisions(single_rule) for single_rule in
                self._split_and_rename_rule(rule)]

    def _resolve_variable_collisions[T1: Rule | _QuestionRule](self, rule: T1) -> T1:
        """Rename variables in a rule to avoid collisions."""
        var_map: dict[str, Variable] = {}  # 仅作为临时工作区
        new_head = self._generate_renamed_item(rule.head, var_map)
        new_body = self._generate_renamed_item(rule.body, var_map)

        return cast("T1", rule.replace(head=new_head, body=new_body))

    def _split_and_rename_rule[T1: Rule](self, rule: T1) -> Sequence[T1]:
        """Assign names to unnamed rules and split rules into DNF."""
        origin_name = rule.name or f"rule_{self.rule_suffix}"
        self.rule_suffix += 1

        rule_spliter = RuleSafetyProcesser()  # TODO: 这里每次创建一个TODO还是有点奇怪
        splited_rules = rule_spliter.split_rule_and_process_safety(rule)

        if len(splited_rules) == 1:
            return [splited_rules[0].replace(name=origin_name)]

        result = []
        for idx, single_rule in enumerate(splited_rules):
            new_name = f"{origin_name}_{idx + 1}"
            result.append(single_rule.replace(name=new_name))

        return result

    @overload
    def _generate_renamed_item(self, item: Variable, var_map: dict[str, Variable]) -> Variable: ...
    @overload
    def _generate_renamed_item(self, item: Constant, var_map: dict[str, Variable]) -> Constant: ...
    @overload
    def _generate_renamed_item(self, item: CompoundTerm, var_map: dict[str, Variable]) -> CompoundTerm: ...
    @overload
    def _generate_renamed_item(self, item: Assertion, var_map: dict[str, Variable]) -> Assertion: ...
    @overload
    def _generate_renamed_item(self, item: Formula, var_map: dict[str, Variable]) -> Formula: ...
    @overload
    def _generate_renamed_item(self, item: None, var_map: dict[str, Variable]) -> None: ...

    def _generate_renamed_item(self,
                               item: Formula | Assertion | CompoundTerm | Variable | Constant | None,
                               var_map: dict[str, Variable]) -> (
            Formula | Assertion | CompoundTerm | Variable | Constant | None):
        if isinstance(item, Variable):
            if item.symbol not in var_map:
                new_name = f"_v{self._var_counter}"
                self._var_counter += 1
                var_map[item.symbol] = item.create_renamed_variable(new_name)
                # risk: 两条规则里如果都用到了x，然后使用者使用了同一个x的instance，
                # 我们是否有理由认为它俩是一个x？（我认为不合理，所以我们应当不管instance，而只是取它的value本身。但这样会不会给使用者带来困惑？
                # 他可能刻意用了同一个instance，甚至在里面存了一些自定义的信息。但我们可能是直接丢弃然后换个了新名字新地址）
            return var_map[item.symbol]

        if isinstance(item, Constant):
            return item

        if isinstance(item, CompoundTerm):
            return CompoundTerm.from_parts(
                item.operator,
                tuple(self._generate_renamed_item(arg, var_map) for arg in item.arguments),
            )

        if isinstance(item, Assertion):
            return Assertion.from_parts(
                self._generate_renamed_item(item.lhs, var_map),
                self._generate_renamed_item(item.rhs, var_map),
            )

        if isinstance(item, Formula):
            return Formula(
                self._generate_renamed_item(item.formula_left, var_map),
                item.connective,
                self._generate_renamed_item(item.formula_right, var_map),
            )

        if item is None:
            return None

        raise ValueError(f"Unknown item type: {type(item)}")

    def __str__(self) -> str:
        rule_count = len(self.rules)

        show_topn = 5

        # 展示前5条规则
        preview_rules = ', '.join(str(rule) for rule in list(self.rules)[:5])
        if rule_count > show_topn:
            preview_rules += "..."

        return f"RuleBase with {rule_count} rules. First 5 rules: [{preview_rules}]"



# 导入所有term strategies
import importlib
import logging
import pathlib

from kele.control.registry import get_term_strategy_class

current_dir = pathlib.Path(__file__).resolve().parent
package_name = __package__ or current_dir.name

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

for filename in current_dir.iterdir():
    if filename.suffix == '.py' and filename.stem.endswith('_strategy') and filename.stem.startswith('_'):
        module_name = filename.stem
        logger.info('successfully imported module: "%s"', module_name)
        try:
            module = importlib.import_module(f'{package_name}.{module_name}')
        except ImportError:
            logger.exception('Failed to import %s', module_name)
            continue

__all__ = ["get_term_strategy_class"]



from ._assertion import _AssertionNode
from ._conn import _ConnectiveNode
from ._op import _OperatorNode
from ._root import _RootNode
from ._rule import _QuestionRuleNode, _RuleNode
from ._term import _BuildTerm, _ConstantNode, _FlatCompoundTermNode, _TermNode, _VariableNode
from ._tftable import TfTables
from ._tupletable import _TupleTable

__all__ = [
           'TfTables',
           '_AssertionNode',
           '_BuildTerm',
           '_ConnectiveNode',
           '_ConstantNode',
           '_FlatCompoundTermNode',
           '_OperatorNode',
           '_QuestionRuleNode',
           '_RootNode',
           '_RuleNode',
           '_TermNode',
           '_TupleTable',
           '_VariableNode',
]



from kele.syntax import Rule

from .builtin_facts import example_assertion_1, example_formula_1

# Example Rules
example_rule_1 = Rule(
    head=example_assertion_1,
    body=example_assertion_1,
    priority=0.1,
)
example_rule_2 = Rule(
    head=example_assertion_1,
    body=example_formula_1,
    priority=0.5,
)



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from kele.control.grounding_selector._selector_utils import (
    _unify_ground_terms_from_rules,
    _unify_into_terms,
)

from ._term_strategies import get_term_strategy_class

if TYPE_CHECKING:
    from collections.abc import Sequence

    from kele.config import Config
    from kele.equality import Equivalence
    from kele.syntax import FACT_TYPE, GROUNDED_TYPE_FOR_UNIFICATION, Rule

    from ._term_strategies.strategy_protocol import Feedback


class GroundingFlatTermWithWildCardSelector:  # 此时是FlatTerm-level grounding，所以进strategy不止是Term。但是strategy本身是Term和
    # FlatTerm均合适的，所以strategy自己的标准仍然得是TERM_TYPE。
    # 此外，目前通配符和flat term无法分割，所以命名强调了wild card
    """
    对外统一入口，内部委托给策略实现。
    """
    def __init__(self,
                 equivalence: Equivalence,
                 args: Config) -> None:

        strategy_cls = get_term_strategy_class(args.strategy.grounding_term_strategy)
        self._strategy = strategy_cls()
        self._equivalence = equivalence
        self._args = args

    def next_terms(self, rule: Rule) -> list[GROUNDED_TYPE_FOR_UNIFICATION]:
        """为给定规则选择候选事实/term 用于 grounding。"""
        init_terms = list(self._strategy.select_next(rule))

        if self._args.run.semi_eval_with_equality:
            selected_terms: set[GROUNDED_TYPE_FOR_UNIFICATION] = set()

            for t in init_terms:
                if t not in selected_terms:
                    equiv_terms = self._equivalence.get_related_item(t)
                    selected_terms |= set(equiv_terms)  # TODO: 这里后面考虑是否把_unify_into_flat_terms挪到term selector

                    # 在semi-evaluation中纳入等价类其余terms，这是由
                    # 断言逻辑带来的内嵌、强制等词公理，和term-level grounding二者共同决定的算法，因为不放置于具体的Strategy中，而是作为
                    # Selector自身的行为。
            return list(selected_terms)

        return list(set(init_terms))

    def active_terms_count(self) -> int | None:
        """返回当前候选 term 数量，若策略不支持则返回 None。"""
        terms = getattr(self._strategy, "_terms", None)
        if terms is None:
            return None
        return len(terms)

    def reset(self) -> None:
        """重置选择器"""
        self._strategy.reset()

    def update_terms(self,
                     terms: Sequence[GROUNDED_TYPE_FOR_UNIFICATION] | None = None,
                     facts: Sequence[FACT_TYPE] | None = None) -> None:
        """
        更新事实集合，并同步给当前策略。
        :raise: ValueError: 没有可选terms_or_facts时报错
        """

        if terms:
            fact_terms = terms
        elif facts:
            fact_terms = list(itertools.chain.from_iterable(_unify_into_terms(f) for f in facts))  # FIXME: 更进一步地，其实只有Assertion
            # 能触发等式公理
        else:
            raise ValueError("terms or facts cannot be empty")

        # 不应有action operator的term。fact_terms = [f for f in fact_terms if f.operator.implement_func is not None]
        self._strategy.add_terms(fact_terms)

    def update_terms_from_rules(self, rules: Sequence[Rule]) -> None:
        """
        将规则/问题中无变量的term加入候选表。
        """
        rule_terms = _unify_ground_terms_from_rules(rules)
        if not rule_terms:
            return

        self._strategy.add_terms(rule_terms)

    def send_feedback(self, feedback: Feedback) -> None:
        """把一次选择后的反馈转发给当前策略"""
        self._strategy.on_feedback(feedback)



"""本体库的导入文件"""

import importlib.util
from pathlib import Path
from types import ModuleType


def _py_auto_load(filename: Path) -> ModuleType:
    """
    从给定的文件路径动态加载一个 Python 模块。

    :param filename: Python 文件路径。
    :type filename: Path
    :return: 加载的模块对象。
    :rtype: types.ModuleType
    :raises ImportError: 如果无法加载模块。
    """
    module_name = filename.stem  # 获取不带扩展名的文件名作为模块名
    spec = importlib.util.spec_from_file_location(module_name, str(filename))
    if spec and spec.loader:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    raise ImportError(f"Unable to import module from {filename}")


def load_ontologies(
    concept_dir_or_path: str | Path,
    operator_dir_or_path: str | Path,
) -> None:
    """
    从概念目录和算子目录（或文件）加载所有 Python 本体模块。

    :param: concept_dir_or_path: 概念模块的目录或文件路径或yaml文件。
    :param: operator_dir_or_path: 算子模块的目录或文件路径或yaml文件。
    """
    def _collect_files(path: str | Path, suffix: str = '.py') -> list[Path]:
        """
        收集指定路径下所有 Python 文件。

        参数:
            path (str | Path): 文件路径或目录路径。

        返回:
            List[Path]: 所有符合条件的 Python 文件路径。
        """
        path = Path(path)
        if path.is_dir():
            # 返回目录下所有 .py 文件（不包括子目录）
            return sorted(path.glob(f"*{suffix}"))

        if path.is_file() and path.suffix == ".py":
            return [path]

        return []

    # 收集概念和算子模块的 Python 文件
    concept_py_files = _collect_files(concept_dir_or_path)
    operator_py_files = _collect_files(operator_dir_or_path)

    all_modules = []
    for file in concept_py_files + operator_py_files:
        module = _py_auto_load(file)
        all_modules.append(module)

    # TODO: 收集概念和算子模块的 yaml 文件



"""用于callbacks和推理路径的记录"""
from .callback import Callback, CallbackManager, HookMixin
from .grounding_selector import GroundingRuleSelector
from .builtin_hooks import BuiltinHookEnabler, register_assertion_check_hook
from .infer_path import InferencePath
from .registry import register
from .status import (
    InferenceStatus,
    create_executor_manager,
    create_main_loop_manager,
)

__all__ = [
    'BuiltinHookEnabler',
    'Callback',
    'CallbackManager',
    'GroundingRuleSelector',
    'HookMixin',
    'InferencePath',
    'InferenceStatus',
    'create_executor_manager',
    'create_main_loop_manager',
    'register',
    'register_assertion_check_hook',
]



"""grounding相关的选择器"""
from kele.control.registry import register

from .rule_selector import GroundingRuleSelector
from .term_selector import GroundingFlatTermWithWildCardSelector

__all__ = [
    "GroundingFlatTermWithWildCardSelector",
    "GroundingRuleSelector",
    "register",
]



# ruff: noqa: PLR6301

"""
这个模块提供四则运算的语法糖，使用 Lark 语法解析器。
请注意每个CompoundTerm都必须用括号括起来，包括但不限于嵌套、原子的CompoundTerm等
支持的符号包括：
- 数字（整数或浮点数）
- 变量名（[A-Za-z_]\\w*）
- 括号（支持嵌套）
- 二元操作符：+、-、*、/
- 一元操作符：+、-
- 方程式：e_term "=" e_term
解析后会将语法树转换为推理引擎的syntax：CompoundTerm、Constant、Variable等
支持的op和concept置于builtin_base文件夹，例如 arithmetic_plus_op、arithmetic_minus_op 等。
"""

from __future__ import annotations

from typing import Final

from lark import Lark, Token, Transformer

from kele.knowledge_bases.builtin_base.builtin_concepts import (
    COMPLEX_NUMBER_CONCEPT,
)
from kele.knowledge_bases.builtin_base.builtin_operators import (
    arithmetic_divide_op,
    arithmetic_minus_op,
    arithmetic_negate_op,
    arithmetic_plus_op,
    arithmetic_times_op,
    get_arithmetic_equation_op,
)
from kele.syntax import TERM_TYPE, Assertion, CompoundTerm, Constant, Formula, Variable

GRAMMAR: Final[str] = r"""
// 顶层：（方程中的）等式 | 带括号表达式 | 裸数字
start: equation
     | pexpr                 -> top_paren_expr
     | number

// （方程中的）等式：两侧允许 带括号表达式 或 裸数字
equation: e_term "=" e_term   -> equation
e_term: pexpr | number

// 带括号表达式：目前是四则运算 + 一元正负，允许嵌套
pexpr: "(" expr ")"          -> paren
     | "(" pexpr ")"         -> nested_paren

// 四则运算
?expr: expr "+" term         -> add
     | expr "-" term         -> sub
     | expr "*" term         -> mul
     | expr "/" term         -> div
     | "-" expr              -> neg
     | "+" expr              -> pos
     | term

?term: number
       | symbol
       | pexpr               // 支持嵌套括号

number: NUMBER
symbol: NAME

NAME: /[A-Za-z_]\w*/
%import common.NUMBER
%ignore /[ \t]+/
"""


class ToSyntax(Transformer):  # type: ignore[type-arg]
    """Transform Lark parse trees into AL inference engine terms."""

    # --- Entrypoints & helpers -------------------------------------------------

    def start(self, items: list[TERM_TYPE | Assertion | Formula]) -> TERM_TYPE | Assertion | Formula:
        """Return the single top-level term."""
        return items[0]

    def e_term(self, items: list[TERM_TYPE]) -> TERM_TYPE:
        """Return an equation-side term unchanged."""
        return items[0]

    # --- Parentheses handling --------------------------------------------------

    def top_paren_expr(self, items: list[TERM_TYPE]) -> TERM_TYPE:
        """Alias for a top-level parenthesized expression."""
        return items[0]

    def paren(self, items: list[TERM_TYPE]) -> TERM_TYPE:
        """Elide a single layer of parentheses."""
        return items[0]

    def nested_paren(self, items: list[TERM_TYPE]) -> TERM_TYPE:
        """Elide nested parentheses."""
        return items[0]

    # --- Atoms -----------------------------------------------------------------

    def number(self, items: list[Token]) -> Constant:
        """Convert a numeric token into a :class:`Constant`."""
        s = str(items[0])
        v = float(s) if ("." in s or "e" in s or "E" in s) else int(s)
        return Constant(v, COMPLEX_NUMBER_CONCEPT)

    def name(self, items: list[Token]) -> Variable:
        """Convert an identifier token into a :class:`Variable`."""
        return Variable(str(items[0]))

    # --- Unary ops -------------------------------------------------------------

    def neg(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build a unary negation term."""
        return CompoundTerm.from_parts(arithmetic_negate_op, [items[0]])

    def pos(self, items: list[CompoundTerm]) -> TERM_TYPE:
        """Return the inner term for unary plus (no-op)."""
        return items[0]

    # --- Binary ops ------------------------------------------------------------

    def add(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build an addition term."""
        left, right = items
        return CompoundTerm.from_parts(arithmetic_plus_op, [left, right])

    def sub(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build a subtraction term."""
        left, right = items
        return CompoundTerm.from_parts(arithmetic_minus_op, [left, right])

    def mul(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build a multiplication term."""
        left, right = items
        return CompoundTerm.from_parts(arithmetic_times_op, [left, right])

    def div(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build a division term."""
        left, right = items
        return CompoundTerm.from_parts(arithmetic_divide_op, [left, right])

    # --- Equation --------------------------------------------------------------

    def equation(self, items: list[CompoundTerm]) -> CompoundTerm:
        """Build an equality term for the left/right sides."""
        left, right = items
        return CompoundTerm.from_parts(get_arithmetic_equation_op, [left, right])


_parser: Lark = Lark(GRAMMAR, parser="lalr", maybe_placeholders=False)
_to_syntax: ToSyntax = ToSyntax()


class SyntacticSugar:
    """Callable facade that delegates to :func:`parse_term`."""

    def __call__(self, input_str: str) -> TERM_TYPE | Assertion | Formula:
        """Parse ``input_str`` and return a transformed term.

        Parameters
        ----------
        input_str
            The input string to parse.

        Returns
        -------
        Term
            The transformed term produced by :func:`parse_term`.

        Raises
        ------
        SyntaxError
            If parsing fails for any reason.
        """
        try:
            return _to_syntax.transform(_parser.parse(s))  # type: ignore[no-any-return]
        except Exception as e:
            raise SyntaxError(f"Parsing failed: {e}\n") from e


syntactic_sugar: Final[SyntacticSugar] = SyntacticSugar()


if __name__ == "__main__":
    cases: list[str] = [
        "(1+2)=3",
        "(x*2+3)",
        "1",
        "((1)+(2))=((3))",  # 多余括号仍可
        "((1)+(2))=3",
        "((1+2))=3",
        "(1+(2*3))=7",
        "(-(1+2))",  # 顶层表达式需括号
        "(1+x)",  # 所有的term都要求有括号
        "(1+(-x))",
        "(((1+x)))",
    ]
    for s in cases:
        print(s, "=>", syntactic_sugar(s))  # noqa: T201



from __future__ import annotations

import logging
import warnings
from collections import deque
from typing import TYPE_CHECKING, Literal

from pyvis.network import Network

from kele.syntax import FACT_TYPE, Assertion, Formula

if TYPE_CHECKING:
    from collections.abc import Sequence

    from kele.config import RunControlConfig
    from kele.equality import Equivalence
    from kele.syntax import Rule

logger = logging.getLogger(__name__)


# 单个推理步：记录一个事实由哪条规则得到，以及它与前后事实的连接
class FactStep:
    """与上游/下游的事实联系的封装，记录事实来源用的"""
    def __init__(self, content: FACT_TYPE, infer_step: Rule | tuple[Assertion, ...] | None,
                fact_type: Literal['premise', 'equivalence', 'rule_infer']) -> None:
        # 当前仅记录等价类推导的“存在性”，不追溯具体等价链路
        # TODO: 可扩展记录推理深度或来源解释
        self.fact_type: Literal['premise', 'equivalence', 'rule_infer'] = fact_type  # 事实的类型
        self.content: FACT_TYPE = content        # 实例化后的事实
        self.infer_step: Rule | tuple[Assertion, ...] | None = infer_step  # 派生该事实的规则，若由等价关系/同余闭包推导则为tuple，
        # 若为前提事实则为 None
        self._next_facts: list[FactStep] = []   # 由当前事实推演出的下游事实
        self._prev_facts: list[FactStep] = []   # 支撑当前事实的上游事实

    def add_next(self, fact: FactStep) -> None:
        """将事实与它帮助推导的下游事实联系起来"""
        self._next_facts.append(fact)

    def add_prev(self, fact: FactStep) -> None:
        """将事实连接到支持它的上游事实"""
        self._prev_facts.append(fact)

    @property
    def next(self) -> tuple[FactStep, ...]:
        """下游事实"""
        return tuple(self._next_facts)

    @property
    def prev(self) -> tuple[FactStep, ...]:
        """上游事实"""
        return tuple(self._prev_facts)

    @property
    def step_name(self) -> str:
        """
        FactStep的名称，用于打印
        """
        if self.fact_type == 'premise':
            return f"无前提事实：{self.content !s}"
        if self.fact_type == 'equivalence':
            return f"等价推出事实：{self.content !s}"
        return f"规则推导：{self.infer_step !s} 新事实({self.content !s})"

    def __repr__(self) -> str:  # pragma: no cover
        rule_name = getattr(self.infer_step, "name", None)
        return f"FactStep({self.content}, rule={rule_name})"

    def __hash__(self) -> int:
        return hash((self.content, self.infer_step))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FactStep) and self.content == other.content and self.infer_step == other.infer_step


class InferencePath:
    """
        存储推理图：
            1. forward  : (antecedent_fact, rule)  -> [consequent_facts]
            2. reverse  : (consequent_fact, rule)  -> [antecedent_facts]
    """
    def __init__(self, args: RunControlConfig, equivalence: Equivalence) -> None:
        self._args = args

        self.fact_factstep_pool: dict[FACT_TYPE, FactStep] = {}
        self.terminal_step: FactStep | None = None  # 记录最后的终点fact，通常是question对应的fact
        self.equivalence: Equivalence = equivalence
        self.initial_facts: set[FACT_TYPE] = set()

        self._fact_counter = 1
        self._step_counter = 1
        self.fact_factid_map: dict[FACT_TYPE, str] = {}
        self.step_stepid_map: dict[str, str] = {}

    def _add_initial_facts(self, facts: Sequence[FACT_TYPE] | FACT_TYPE) -> None:
        """
        添加初始事实
        """
        if isinstance(facts, FACT_TYPE):
            facts = [facts]

        for fact in facts:
            if isinstance(fact, Assertion):
                self.initial_facts.add(fact)
                self.initial_facts.add(self._reverse_fact(fact))
            else:
                self.initial_facts.add(fact)

    def _is_validate_none_premise_assertion(self, fact: Assertion) -> bool:
        """
        检查一个fact是否是一个合法的前提为None的Assertion
        以下两种情况前提为None
        1、fact出现在initial_fact里面
        2、左右显然相等
        """
        return fact in self.initial_facts or fact.lhs == fact.rhs or fact.is_action_assertion

    def _query_equiv_step(self, fact: FACT_TYPE) -> FactStep:
        """
        获取一个fact的推理路径，总共由三个可能：
        1. 它是由等价关系推出来的
        2. 它是由规则推出来的
        3. 它是一个前提事实
        最后都会返回一个FactStep

        :param fact: 待检查的事实
        :type fact: FACT_TYPE
        :raises RuntimeError: 若等价关系处理器未设置
        :return: 若fact是由等价关系推出来的，则返回等价关系的FactStep，否则返回None
        :rtype: FactStep
        """
        if fact in self.fact_factstep_pool:
            # fact 已记录过推理路径，直接复用
            return self.fact_factstep_pool[fact]
        if isinstance(fact, Assertion) and self._is_validate_none_premise_assertion(fact):
            # fact是一个前提事实
            return FactStep(fact, None, 'premise')
        # Assertion的factstep需要考虑是否是等价关系推导出来的，但是Formula类型的（实质只可能为NOT Assertion）则不需要
        if isinstance(fact, Formula) and isinstance(fact.formula_left, Assertion) and fact.connective == 'NOT':
            # NOT Assertion类型的Fact不需要考虑等价关系
            # 在正常情况下，它自然是成立的前提事实，否则是不可能推理出结果的
            return FactStep(fact, None, 'premise')
        if isinstance(fact, Formula):
            raise TypeError(
                "Rule premises cannot contain connectives other than AND and NOT. "
                "This error may come from CNF_convert."
            )
        if self.equivalence is None:
            raise RuntimeError(
                "Equivalence handler is not set; cannot properly record inference paths for equivalence facts."
            )
        if self.equivalence.query_equivalence(fact):
            fact_step = FactStep(fact, None, 'equivalence')  # HACK：暂时不详细处理等价关系推出的事实，
            # 后续需要获取等价关系的解释
            self.fact_factstep_pool[fact] = fact_step
            self.fact_factstep_pool[self._reverse_fact(fact)] = fact_step  # 对称事实也要记录进去
            return fact_step

        raise ValueError(f"Fact {fact!s} is not true; cannot record inference path.")

    @staticmethod
    def _reverse_fact(fact: FACT_TYPE) -> FACT_TYPE:
        if isinstance(fact, Assertion):
            return Assertion.from_parts(fact.rhs, fact.lhs)
        return fact

    def add_infer_edge(self,
                       consequent: FACT_TYPE,  # FIXME: 这里得缩减为Assertion
                       antecedents: list[FACT_TYPE] | None = None,
                       grounded_rule: Rule | None = None,
                    ) -> None:
        """
        录入一条推理边：多前提 → 单结论
        :param antecedents: 对应某条rule的前提，不过已经实例化过了
        :param consequent: 对应规则后件的实例化结果
        :param grounded_rule: 触发推理的规则
        :return: None
        """
        if antecedents is None:
            return self._add_initial_facts(consequent)

        if consequent in self.fact_factstep_pool:
            # 事实已经存在，推理路径默认保留一条即可
            # TODO: 可选保留多条推理路径
            return None
        # 记录结论的推理路径
        conse_step = FactStep(consequent, grounded_rule, 'rule_infer')

        self.fact_factstep_pool[consequent] = conse_step
        self.fact_factstep_pool[self._reverse_fact(consequent)] = conse_step  # 对称事实也要记录进去
        for fact in antecedents:
            factstep = self._query_equiv_step(fact)
            self.fact_factstep_pool[fact] = factstep
            self.fact_factstep_pool[self._reverse_fact(fact)] = factstep  # 对称事实也要记录进去

            factstep.add_next(conse_step)
            conse_step.add_prev(factstep)
        return None

    def add_terminal_status(self, termnial_fact: FACT_TYPE) -> None:
        """记录终点事实"""
        try:
            self.terminal_step = self._query_equiv_step(termnial_fact)  # termnimal_step也要考虑由等价关系推出的可能
        except ValueError:
            warnings.warn(f"Terminal fact {termnial_fact!s} is trivially true.", stacklevel=1)
            self.terminal_step = None

    @staticmethod
    def _print_log_info(prev_fact_steps: list[FactStep], infer_path: deque[FactStep], terminal_fact: FACT_TYPE) -> None:
        logger.info("================Premise facts:=================")
        for prev_fact_counter, fact_step in enumerate(prev_fact_steps):
            logger.info("%d. %s", prev_fact_counter + 1, fact_step.step_name)  # FIXME: 这里的注释有点奇怪，一个数字
            # 一个name。博洋改到这里的时候留意一下，反正你的infer最近要动，我就不细究了
        logger.info("================Inference path:=================")
        for infer_fact_counter, fact_step in enumerate(infer_path):
            logger.info("step %d: %s", infer_fact_counter + 1, fact_step.step_name)
        logger.info("================Terminal fact:=================")
        logger.info("Terminal fact: %s", terminal_fact)

    def get_infer_graph(self, terminal_fact: FACT_TYPE | None = None) -> tuple[list[FactStep], FACT_TYPE | None]:
        """
        获得推理路径
        :param terminal_fact: 推理的终点事实，默认是question对应的fact
        :return: 推理路径，终点事实
        """
        if not self._args.trace:
            warnings.warn("Inference path tracing is disabled; cannot print inference path.", stacklevel=5)
            return [], None
        terminal_step = self.terminal_step if terminal_fact is None else self._query_equiv_step(terminal_fact)

        infered: set[FactStep] = set()
        if terminal_step is not None:
            infer_path: deque[FactStep] = deque()
            prev_fact_steps: list[FactStep] = []
            cur_fact_queue = deque([terminal_step])
            while cur_fact_queue:
                cur_fact_step = cur_fact_queue.popleft()
                if cur_fact_step in infered:  # 推出的事实不再重复推导
                    continue
                infered.add(cur_fact_step)

                if cur_fact_step.fact_type != 'premise':
                    infer_path.appendleft(cur_fact_step)
                else:
                    # 前提事实全部在第一步展示
                    # 这里的前提事实指的是原始Premises中真正被用于推理的那些前提事实
                    prev_fact_steps.append(cur_fact_step)
                if cur_fact_step.infer_step is not None:
                    cur_fact_queue.extend(cur_fact_step.prev)

            self._print_log_info(prev_fact_steps, infer_path, terminal_step.content)
            prev_fact_steps.extend(infer_path)  # 将分开的两个集合合并起来返回，此时顺序已经被确定下来
            return prev_fact_steps, terminal_step.content
        warnings.warn("Inference engine could not derive a result, or the terminal fact is trivially true.", stacklevel=1)
        return [], None

    def _get_fact_id(self, fact: FACT_TYPE, net: Network) -> str:
        if fact not in self.fact_factid_map:
            self.fact_factid_map[fact] = f"fact{self._fact_counter}"
            self._fact_counter += 1
            net.add_node(self.fact_factid_map[fact], label=self.fact_factid_map[fact], title=str(fact))
        return self.fact_factid_map[fact]

    def _get_step_id(self, step_name: str, net: Network) -> str:
        if step_name not in self.step_stepid_map:
            self.step_stepid_map[step_name] = f"step{self._step_counter}"
            self._step_counter += 1
            net.add_node(self.step_stepid_map[step_name], label=self.step_stepid_map[step_name], title=str(step_name), shape="square", color="red")
        return self.step_stepid_map[step_name]

    def gennerate_infer_path_graph(self, infer_path: list[FactStep], terminal_fact: FACT_TYPE | None = None) -> None:
        """
        生成推理路径的图
        :param infer_path: 推理路径
        :param terminal_fact: 推理的终点事实，默认是question对应的fact
        :return: None
        """
        net = Network(height="600px", width="100%", bgcolor="#ffffff", font_color="black")
        self._fact_counter = 1
        self._step_counter = 1
        self.fact_factid_map.clear()
        self.step_stepid_map.clear()
        if terminal_fact is not None:
            self.fact_factid_map[terminal_fact] = "terminal_fact"
            net.add_node("terminal_fact", label="终点事实", shape="star", color="red")

        for fact_step in infer_path:
            cur_fact_id = self._get_fact_id(fact_step.content, net)
            cur_step_id = self._get_step_id(fact_step.step_name, net) if fact_step.infer_step is not None else None
            if cur_step_id is not None:
                for fact in fact_step.prev:
                    prev_fact_id = self._get_fact_id(fact.content, net)
                    net.add_edge(prev_fact_id, cur_step_id, label="前提", color="blue")
                net.add_edge(cur_step_id, cur_fact_id, label="结论", color="red", arrows="to")
            else:
                for nodes in net.nodes:
                    if nodes["id"] == cur_fact_id:
                        nodes["label"] = "无前提事实"
                        nodes["shape"] = "triangle"
                        nodes["color"] = "green"
        net.save_graph("infer_path.html")

    def reset(self) -> None:
        """重置推理路径"""
        self.fact_factstep_pool.clear()
        self.terminal_step = None
        self.initial_facts.clear()



from __future__ import annotations

import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, Literal

from kele.config import ExecutorConfig, RunControlConfig
from kele.syntax import Variable, _QuestionRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from kele.equality import Equivalence
    from kele.syntax import FACT_TYPE, CompoundTerm, Constant, Question, Rule, SankuManagementSystem

logger = logging.getLogger(__name__)


class InferenceStatus(Enum):
    """推理过程的终止状态类型"""

    SUCCESS = auto()  # 成功推理出答案
    MAX_STEPS_REACHED = auto()  # 达到最大执行轮次限制
    MAX_ITERATIONS_REACHED = auto()  # 达到最大迭代次数限制（从main循环）
    FIXPOINT_REACHED = auto()  # 达到不动点
    CONFLICT_DETECTED = auto()  # 推理中发现矛盾
    EXTERNALLY_INTERRUPTED = auto()  # 被外部中断（如人为终止）
    NO_MORE_RULES = auto()  # 没有更多实例化规则可以推理
    CONTINUE = auto()  # 当前轮未结束，继续执行

    def log_message(self) -> str:
        """每个状态对应的日志输出信息"""
        return {
            InferenceStatus.SUCCESS: "Successfully inferred the answer.",
            InferenceStatus.MAX_STEPS_REACHED: "Max execution steps reached, terminating.",
            InferenceStatus.MAX_ITERATIONS_REACHED: "Max inference iterations reached, terminating.",
            InferenceStatus.FIXPOINT_REACHED: "Fixpoint reached, terminating.",
            InferenceStatus.CONFLICT_DETECTED: "Conflict detected during execution.",
            InferenceStatus.EXTERNALLY_INTERRUPTED: "Execution was externally interrupted.",
            InferenceStatus.NO_MORE_RULES: "No more grounded rules available for execution.",
            InferenceStatus.CONTINUE: "Execution continues to next round."
        }[self]

    def is_terminal_for_executor(self) -> bool:
        """判断执行器是否应该终止"""
        return self != InferenceStatus.CONTINUE

    def is_terminal_for_main_loop(self) -> bool:
        """判断主循环是否应该终止"""
        return self in {
            InferenceStatus.SUCCESS,
            InferenceStatus.MAX_STEPS_REACHED,
            InferenceStatus.MAX_ITERATIONS_REACHED,
            InferenceStatus.FIXPOINT_REACHED,
            InferenceStatus.EXTERNALLY_INTERRUPTED
        }


class QuerySolutionManager:
    """查询解管理器：负责收集、扩展（等价解）、去重并按需打印查询解。"""

    def __init__(self, args: RunControlConfig) -> None:
        self.collected_solutions: list[Mapping[Variable, Constant | CompoundTerm]] = []
        self.args = args

        # 统计“已记录的唯一解”数量（包含等价扩展后的解）
        self._solution_count: int = 0

        # 记录/输出去重
        self._seen_solution_keys: set[tuple[tuple[str, str], ...]] = set()  # 去重：已出现过的解（含等价扩展）
        self._printed_solution_keys: set[tuple[tuple[str, str], ...]] = set()  # 记录已输出的解，避免重复打印
        self._pending_solutions: list[Mapping[Variable, Constant | CompoundTerm]] = []  # 待输出：新增未打印解（含等价扩展）

    def reset(self) -> None:
        """清空已收集的解、去重状态和计数。"""
        self.collected_solutions.clear()
        self._solution_count = 0
        self._seen_solution_keys.clear()
        self._printed_solution_keys.clear()
        self._pending_solutions.clear()

    # -----------------------------
    # add（仅记录，不打印）
    # -----------------------------
    def add_solution(
        self,
        solution: Mapping[Variable, Constant | CompoundTerm],
        *,
        equivalence: Equivalence | None = None,
    ) -> None:
        """记录 1 个 solution，并对其每个绑定值做等价扩展，只合入新的 solutions。

        返回：本次新增（去重后）记录的解数量。
        """
        self._record_with_equivalents(solution, equivalence=equivalence)

    def add_solutions(
        self,
        solutions: Sequence[Mapping[Variable, Constant | CompoundTerm]],
        *,
        equivalence: Equivalence | None = None,
    ) -> None:
        """批量记录 solutions（并对每个解做等价扩展）。"""
        for s in solutions:
            self.add_solution(s, equivalence=equivalence)

    def has_pending_solutions(self) -> bool:
        """是否存在尚未打印的新解。"""
        return bool(self._pending_solutions)

    # -----------------------------
    # print（仅输出，不新增记录）
    # -----------------------------
    def print_pending(
        self,
        question: Question,
        question_rule: _QuestionRule,
    ) -> bool:
        """打印自上次调用以来新加入的解（含等价扩展），并根据配置决定是否终止推理。

        :return: 是否应该终止推理
        """
        if not self._pending_solutions:
            return False

        question_str = ", ".join(str(q) for q in question.question)

        while self._pending_solutions:
            # HACK: 使用 pop() 逆序输出新增解；当前没有需求时不特地排序或引入 deque 优化。
            combination = self._pending_solutions.pop()
            key = self._solution_key(combination)

            # pending 理论上都未打印过，但这里仍做一次防御性处理
            if key in self._printed_solution_keys:
                continue

            self._printed_solution_keys.add(key)

            if combination:
                # 输出时除了 matcher 找到的绑定，也包含等价扩展后的绑定。
                binding_str = self._display_binding(combination)
                logger.result("Query solution: %s variable bindings: %s", question_str, binding_str)  # type: ignore[attr-defined]
            else:
                logger.result("Query solution: %s", question_str)  # type: ignore[attr-defined]
                if not question_rule.free_variables:
                    return True

            if self._should_terminate(self.args.interactive_query_mode):
                return True

        self._pending_solutions.clear()
        return False

    # -----------------------------
    # backward compatible wrapper
    # -----------------------------
    def print_and_collect(
        self,
        question: Question,
        solutions: list[Mapping[Variable, Constant | CompoundTerm]],
        question_rule: _QuestionRule,
        *,
        equivalence: Equivalence | None = None,
    ) -> bool:
        """兼容旧接口：先 add 再 print。"""
        self.add_solutions(solutions, equivalence=equivalence)
        return self.print_pending(question, question_rule)

    def _record_with_equivalents(
        self,
        solution: Mapping[Variable, Constant | CompoundTerm],
        *,
        equivalence: Equivalence | None,
    ) -> None:
        """记录 solution 及其等价扩展解（去重），不做打印。"""
        self._register_solution(solution)

        # interactive_query_mode=first 时仅输出第一个解，不做等价扩展（add_solutions 同样走该逻辑）。
        if self.args.interactive_query_mode == "first":
            self._solution_count = len(self._seen_solution_keys)
            return

        # 对每个绑定值做等价扩展：逐变量替换。
        if equivalence is not None and solution:
            items = sorted(solution.items(), key=lambda kv: kv[0].display_name)
            for var, term in items:
                for equiv_term in equivalence.get_equiv_item(term):
                    expanded = dict(solution)
                    expanded[var] = equiv_term  # 这里开销会比较大，不过早晚得承担。需要优化的话可以挪到打印solution时候再干
                    self._register_solution(expanded)

        self._solution_count = len(self._seen_solution_keys)

    def _register_solution(self, solution: Mapping[Variable, Constant | CompoundTerm]) -> bool:
        """将解注册到 seen/pending/collected（去重后）。"""
        key = self._solution_key(solution)
        if key in self._seen_solution_keys:
            return False

        self._seen_solution_keys.add(key)

        # 只把“从未打印过”的解放入 pending
        if key not in self._printed_solution_keys:
            self._pending_solutions.append(dict(solution))

        if self.args.save_solutions:
            self.collected_solutions.append(dict(solution))

        return True

    @staticmethod
    def _solution_key(solution: Mapping[Variable, Constant | CompoundTerm]) -> tuple[tuple[str, str], ...]:
        """用于去重的稳定 key：按变量 display_name 排序后，取 (var_display_name, term_str)。"""
        if not solution:
            return ()
        items = sorted(solution.items(), key=lambda kv: kv[0].display_name)
        return tuple((var.display_name, str(term)) for var, term in items)

    @staticmethod
    def _display_binding(combination: Mapping[Variable, Constant | CompoundTerm]) -> str:
        entries: list[str] = []
        for var, term in sorted(combination.items(), key=lambda kv: kv[0].display_name):
            entries.append(f"{var.display_name}={term}")
        return ", ".join(entries)

    @staticmethod
    def _should_terminate(mode: Literal["interactive", "first", "all"]) -> bool:
        if mode == "first":
            logger.info("Configured to output only the first query solution; stopping inference.")
            return True
        if mode == "all":
            return False

        user_input = input("发现解，输入 ';' 并回车继续本次推理；输入其他任意键并回车将终止本次推理: ").strip()  # TODO: input不太适合后续扩展
        if user_input != ';':
            logger.info("用户选择终止推理。")
            return True
        return False

    def get_all_solutions(self) -> list[Mapping[Variable, Constant | CompoundTerm]]:
        """获取所有的solutions"""
        return self.collected_solutions

    def get_solution_count(self) -> int:
        """获取solution的数目"""
        return self._solution_count


class StatusChecker:
    """通用状态检查逻辑"""

    def __init__(self, equivalence: Equivalence, sk_system_handler: SankuManagementSystem):
        self.equivalence = equivalence
        self.sk_system_handler = sk_system_handler

    def check_conflict(self, new_facts: list[FACT_TYPE]) -> bool:
        """检查是否发生了矛盾。!!! 注意这里传入的是new_facts"""
        return self._has_conflict_occurred(new_facts)

    @staticmethod
    def _has_conflict_occurred(new_facts: list[FACT_TYPE]) -> bool:
        """检查是否发生了矛盾"""
        # TODO: 实现矛盾检测逻辑
        return False


class MainLoopManager:
    """主循环管理器"""

    def __init__(self, status_checker: StatusChecker, args: RunControlConfig | None = None):
        self.status_checker = status_checker
        self.args = args or RunControlConfig()
        self._current_iteration = 0

        self.normal_rule_activated: dict[Rule, bool] = {}  # 如果某条rule上一轮有新事实生成，则True；反之False
        self._true_count = -1

    def check_status(self,
                     current_facts: list[FACT_TYPE],
                     question: Question) -> InferenceStatus:
        """检查主循环状态"""
        # 先检查迭代次数
        if self._current_iteration >= self.args.iteration_limit:
            return InferenceStatus.MAX_ITERATIONS_REACHED

        if self._true_count == 0:
            return InferenceStatus.FIXPOINT_REACHED

        if self._current_iteration == 0 and question.question and all(q in current_facts for q in question.question):
            # FIXME: 这里的判断可能需要分的更细致
            return InferenceStatus.SUCCESS

        return InferenceStatus.CONTINUE

    def next_iteration(self) -> None:
        """进入下一轮迭代"""
        self._current_iteration += 1

    def reset(self) -> None:
        """重置计数"""
        self._current_iteration = 0

    def initial_manager(self, normal_rules: Sequence[Rule] | None = None, *, resume: bool = False) -> None:
        """
        修改current rules, 为当前一个question的推理做准备
        :raise ValueError: 如果待推理问题不变，仅中止引擎并重新推理时，认为不需要修改rules（其他各处也应调整）
        """
        if normal_rules is None:
            if resume:
                self.normal_rule_activated = dict.fromkeys(self.normal_rule_activated, True)  # 继续推理当前时，使用之前存储好的rule
                # 但是清空之前的_activated记录（因为可能有新事实变化）
            else:
                raise ValueError("normal_rules is None")
        else:
            self.normal_rule_activated = dict.fromkeys(normal_rules, True)  # 如果某条rule上一轮有新事实生成，则True；反之False
        self._true_count = len(self.normal_rule_activated)

    @property
    def iteration(self) -> int:
        """获取当前迭代次数"""
        return self._current_iteration

    def update_normal_rule_activation(self, new_facts: list[FACT_TYPE], used_rule: Rule) -> None:
        """每条规则完成推理后，使用本函数更新main_manager，用于判断是否所有的规则同时达到了不动点"""
        activated = bool(new_facts)

        old_value = self.normal_rule_activated.get(used_rule)
        self.normal_rule_activated[used_rule] = activated

        if old_value and not activated:
            self._true_count -= 1
        elif not old_value and activated:
            self._true_count += 1

    def is_at_fixpoint(self) -> bool:
        """是否所有 normal rule 都已不再产生新事实（不动点，仅看 normal rules）"""
        return self._true_count == 0


class ExecutorManager:
    """执行器管理器"""

    def __init__(self, status_checker: StatusChecker, solution_manager: QuerySolutionManager,
                 args: ExecutorConfig):
        self.status_checker = status_checker
        self.args = args
        self._current_step = 1
        self.solution_manager = solution_manager

    def check_status(self, new_facts: list[FACT_TYPE], question: Question,
                     solutions: list[Mapping[Variable, Constant | CompoundTerm]],
                     question_rule: _QuestionRule | None) -> InferenceStatus:
        """检查执行器状态"""
        if self.args.executing_max_steps != -1 and self._current_step >= self.args.executing_max_steps:
            return InferenceStatus.MAX_STEPS_REACHED  # 有解是success，且需要解

        # 检查是否有矛盾
        if self.status_checker.check_conflict(new_facts):  # TODO: 冲突时同步清理 solutions
            return InferenceStatus.CONFLICT_DETECTED

        if question_rule is not None:
            # 1) 记录 matcher 找到的解，并同步记录其等价元素
            if solutions:
                completed_solutions = self._complete_question_solutions(question, solutions)
                self.solution_manager.add_solutions(
                    completed_solutions,
                    equivalence=self.status_checker.equivalence,
                )

            # 2) 仅打印尚未输出过的新解
            if self.solution_manager.has_pending_solutions():
                terminate_sign = self.solution_manager.print_pending(question, question_rule)
                if terminate_sign:
                    return InferenceStatus.SUCCESS

        return InferenceStatus.CONTINUE

    @staticmethod
    def _complete_question_solutions(
        question: Question,
        solutions: list[Mapping[Variable, Constant | CompoundTerm]],
    ) -> list[Mapping[Variable, Constant | CompoundTerm]]:
        if not question.free_variables:
            return solutions

        filler = _QuestionRule.QUESTION_VAR_FILLER
        filled_solutions: list[Mapping[Variable, Constant | CompoundTerm]] = []

        for solution in solutions:
            known_vars = [k.display_name for k in solution]
            missing_vars = [var for var in question.free_variables if var.display_name not in known_vars]
            # HACK: 由于question变更为question rule时候进行了重命名（且由于可能被DNF split的缘故，这个重命名似乎是必要的）
            # 系统中question和question rule的Variable instance之间产生了不一致。这里做一层初步修复
            # 这里会遗留一个问题是，solution里面的Variable instance其实是不一样的，只是打印的display name一样
            filled_solutions.append({**solution, **dict.fromkeys(missing_vars, filler)})

        return filled_solutions

    def next_step(self) -> None:
        """执行下一步"""
        self._current_step += 1

    def reset_for_new_inference(self) -> None:
        """为新的推理过程重置步数计数（仅在开始全新推理时调用）"""
        self._current_step = 1

    @property
    def step_num(self) -> int:
        """获取当前步数"""
        return self._current_step


def create_main_loop_manager(equivalence: Equivalence,
                             sk_system_handler: SankuManagementSystem,
                             args: RunControlConfig | None = None) -> MainLoopManager:
    """创建主循环管理器"""
    status_checker = StatusChecker(equivalence, sk_system_handler)
    return MainLoopManager(status_checker, args)


def create_executor_manager(equivalence: Equivalence,
                            sk_system_handler: SankuManagementSystem,
                            solution_manager: QuerySolutionManager,
                            args: ExecutorConfig) -> ExecutorManager:
    """创建执行器管理器"""
    status_checker = StatusChecker(equivalence, sk_system_handler)
    return ExecutorManager(status_checker, solution_manager, args)
